@@ -9,18 +9,22 @@
 #' 
 #' @param expr An expression to be simplified, expr can be
 #' \itemize{
-#'    \item an expression: \code{as.expression(x+x)}
-#'    \item an string: \code{"x+x"}
+#'    \item an expression: \code{expression(x+x)}
+#'    \item a string: \code{"x+x"}
 #'    \item a function: \code{function(x) x+x}
 #'    \item a right hand side of a formula: \code{~x+x}
 #'    \item a language: \code{quote(x+x)}
 #' }
 #' @param env An environment in wich a simplified function is created
-#'  if \code{expr} is a function. This argument is ignored is all other cases.
+#'  if \code{expr} is a function. This argument is ignored in all other cases.
 #' @return A simplified expression. The result is of the same type as
 #'  \code{expr} except for formula, where a language is returned.
 #' @details An environment \code{simplifications} containing simplification rules, is exported in the user namespace.
-Simplify <- function(expr, env=parent.frame())
+Simplify <- function(expr, env=parent.frame()) {
+	te <- try(expr, silent=TRUE)
+	if (inherits(te, "try-error")) {
+		expr <- substitute(expr)
+	}
 	if (is.expression(expr)) {
 		as.expression(Simplify_(expr[[1]]))
 	} else if (is.function(expr)) {
@@ -34,6 +38,8 @@ Simplify <- function(expr, env=parent.frame())
 	} else {
 		Simplify_(expr)
 	}
+}
+
 #' @name format1
 #' @title Wrapper for base::format() function
 # \usage{
@@ -54,24 +60,19 @@ format1 <- function(expr) {
 
 Simplify_ <- function(expr)
 {
-	if (is.unumeric(expr)) {
-		eval(expr)
-	} else if (is.call(expr)) {
-		args <- lapply(expr[-1], Simplify_)
+	if (is.call(expr)) {
+		args <- lapply(as.list(expr)[-1], Simplify_)
+		expr[-1]=args
 		if (all(sapply(args, is.numeric))) {
 			# if all arguments are numeric, evaluate them
-			return(eval(as.call(c(expr[[1]], args))))
+			return(eval(expr))
 		} else {
 			# is there a rule in the table?
 			sym.name <- as.character(expr[[1]])
-			if (class(try(Simplify.rule <-
-					get(sym.name, envir=simplifications,
-					inherits=FALSE), silent=TRUE))
-					!= "try-error") {
-				expr[-1]=args
+			Simplify.rule <- simplifications[[sym.name]]
+			if (!is.null(Simplify.rule)) {
 				return(Simplify.rule(expr))
 			} else {
-				expr[-1]=args
 				return(expr)
 			}
 		}
@@ -102,104 +103,111 @@ Simplify_ <- function(expr)
 	}
 	a <- expr[[2]]
 	b <- expr[[3]]
-	sminus <- FALSE
 	
 	if (a == 0) {
 		return(if (add) b else call("-", b))
 	} else if (b == 0) {
 		return(a)
-	} else if (format1(a) == format1(b) && !add) {
-		return(0)
-	} else if (is.uminus(b)) {
-		b <- b[[2]]
-		add <- !add
-		if (!add && is.uminus(a)) {
-			a <- a[[2]]
-			expr <- call("-", call("+", a, b))
-			sminus <- TRUE
-			add <- TRUE
-		} else {
-			expr <- call(if (add) "+" else "-", a, b)
-		}
+	} else if (add && is.uminus(a) && !is.uminus(b)) {
+		a <- b
+		b <- expr[[2]][[2]]
+		expr <- call("-", a, b)
+	} else if (format1(a) == format1(b)) {
+		return(if (add) call("*", 2, a) else 0)
+	} else if (!is.call(a) && !is.call(b)) {
+		return(expr) # nothing to simplify
 	}
-	# group and simplify identical terms
+	# factorise most repeated terms
 	alc <- Lincomb(a)
 	blc <- Lincomb(b)
-	if (add)
-		lc <- list(co=c(alc$co, blc$co), it=c(alc$it, blc$it))
-	else
-		lc <- list(co=c(alc$co, -blc$co), it=c(alc$it, blc$it))
-	if (sminus)
-		lc$co <- -lc$co # sminus is no more used here after
-	# group all numerics
-	inum <- sapply(lc$it, `==`, 1)
-	nu <- sum(lc$co[inum])
-	# remove numerics from it list
-	lc$it <- lc$it[!inum]
-	lc$co <- lc$co[!inum]
-	# group identical terms
-	itch <- sapply(lc$it, format1)
-	ta <- table(itch) # count of unique entries
-	if (sum(inum) <= 1 && all(ta == 1)) {
-		# nothing to simplify
-		return(expr)
+	if (add) {
+		lc <- c(alc, blc)
+	} else {
+		# inverse sminus in b
+		blc <- lapply(blc, function(it) {it$sminus <- !it$sminus; it})
+		lc <- c(alc, blc)
 	}
-	tsim <- outer(itch, names(ta), `==`)
-	isim <- lapply(seq_len(ncol(tsim)), function(i) which(tsim[,i]))
-	# keep only unique terms
-	irm <- c()
-	for (i in isim) {
-		if (length(i) == 0)
-			next
-		lc$co[i[1]] <- sum(lc$co[i])
-		irm <- c(irm, i[-1])
+	bch <- ta <- tsim <- po <- ilc <- ind <- list()
+	for (cnd in c("num", "den")) {
+		# character bases in num/den
+		bch[[cnd]] <- unlist(lapply(lc, function(it) {lapply(it[[cnd]]$b, format1)}))
+		# powers
+		po[[cnd]] <- do.call(c, lapply(lc, function(it) it[[cnd]]$p), quote=TRUE)
+		# index of the lc term for each bnch
+		ta[[cnd]] <- table(bch[[cnd]])
+		ta[[cnd]] <- ta[[cnd]][ta[[cnd]] > 1] # keep only repeated bases
+		tsim[[cnd]] <- outer(bch[[cnd]], names(ta[[cnd]]), `==`)
+		ilc[[cnd]] <- unlist(lapply(seq_along(lc), function(i) {rep(i, length(lc[[i]][[cnd]]$b))}))
+		# index of the base in a given term (nd) for each bnch
+		ind[[cnd]] <- unlist(lapply(seq_along(lc), function(i) {seq_along(lc[[i]][[cnd]]$b)}))
 	}
-	# remove grouped terms but first
-	if (length(irm)) {
-		lc$co <- lc$co[-irm]
-		lc$it <- lc$it[-irm]
-		itch <- itch[irm]
-	}
-	
-	# remove co==0
-	inul <- lc$co == 0
-	lc$co <- lc$co[!inul]
-	lc$it <- lc$it[!inul]
-	itch <- itch[!inul]
-	if (length(lc$it) == 0) {
-		return(0)
+	# fnd will be the name "num" or "den" where the first factor
+	# will be taken. ond is the "other" name (if fnd=="num", then ond == "den")
+	# we select the cadidate which is most repeated provided that it
+	# has at least one numeric power occurance.
+	taa <- unlist(ta)
+	ota <- order(taa, decreasing=TRUE)
+	ntan <- length(ta$num)
+	fnd <- NA
+	for (i in ota) {
+		cnd <- if (i > ntan) "den" else "num"
+		ita <- i - if (i > ntan) ntan else 0
+		ib <- bch[[cnd]] == names(ta[[cnd]])[ita]
+		if (any(sapply(po[[cnd]], is.numeric))) {
+			fnd <- cnd
+			iit <- which(ib) # the bases equal to factor
+			p_fa <- min(sapply(po[[cnd]][ib], function(p) if (is.numeric(p)) p else NA), na.rm=TRUE)
+			i_lc <- ilc[[cnd]][iit]
+			i_nd <- ind[[cnd]][iit]
+			break
+		}
 	}
 #browser()
-	ipn=list(pos=lc$co > 0, neg=lc$co < 0)
-	for (pn in c("pos", "neg")) {
-		# where abs(co) != 1, replace term by nb_repeat*term
-		i <- ipn[[pn]] & abs(lc$co[ipn[[pn]]]) != 1
-		lc$it[i] <- lapply(which(i), function(ii) {
-			e <- lc$it[[ii]]
-			if (is.call(e) && e[[1]] == as.symbol("/") && e[[2]] == 1)
-				e[[2]] <- abs(lc$co[ii])
-			else
-				e <- Simplify_(call("*", abs(lc$co[ii]), e))
-			return(e)
-		})
+	if (is.na(fnd))
+		return(expr) # nothing to factorize
+	ond <- if (fnd == "num") "den" else "num"
+	# create nd with the first factor
+	fa_nd <- list(num=list(b=list(), p=list()),
+		den=list(b=list(), p=list()),
+		sminus=FALSE, fa=list(num=1, den=1))
+	fa_nd[[fnd]]$b <- lc[[i_lc[1]]][[fnd]]$b[i_nd[1]]
+	fa_nd[[fnd]]$p <- list(p_fa)
+	# decrease p in the lc terms
+	for (i in seq_along(i_lc)) {
+		lc[[i_lc[i]]][[fnd]]$p[[i_nd[i]]] <- Simplify_(call("-", lc[[i_lc[i]]][[fnd]]$p[[i_nd[i]]], p_fa))
 	}
-	if (nu != 0) {
-		lc$it <- c(lc$it, abs(nu))
-		lc$co <- c(lc$co, sign(nu))
-		itch <- c(itch, format1(abs(nu)))
-	}
-	# form final symbolic expression
-	iord <- order(lc$co < 0, itch) # positives first
-	expr <- lc$it[[iord[1]]]
-	sminus <- lc$co[iord[1]] < 0 # all negatives
 	
-	for (i in iord[-1]) {
-		expr <- call(if (sminus || lc$co[i] > 0) "+" else "-", expr, lc$it[[i]])
+	for (cnd in c(fnd, ond)) {
+		# see if other side can provide factors
+		for (i in seq_along(ta[[cnd]])) {
+			if ((cnd == fnd && i == ita) || ta[[fnd]][ita] != ta[[cnd]][i] || any(ilc[[cnd]][tsim[[cnd]][,i]] != i_lc)) {
+				next # no common layout with factor
+			}
+			ib <- bch[[cnd]] == names(ta[[cnd]])[i]
+			# see if it has numeric power
+			if (!any(sapply(po[[cnd]], is.numeric))) {
+				next
+			}
+			iit <- which(ib) # the bases equal to factor
+			p_fa <- min(sapply(po[[cnd]][ib], function(p) if (is.numeric(p)) p else NA), na.rm=TRUE)
+			i_lc <- ilc[[cnd]][iit]
+			i_nd <- ind[[cnd]][iit]
+			fa_nd[[cnd]]$b <- append(fa_nd[[cnd]]$b, lc[[i_lc[1]]][[cnd]]$b[i_nd[1]])
+			fa_nd[[cnd]]$p <- append(fa_nd[[cnd]]$p,
+p_fa)
+			# decrease p in the lc terms
+			for (i in seq_along(i_lc)) {
+				lc[[i_lc[i]]][[cnd]]$p[[i_nd[i]]] <- Simplify_(call("-", lc[[i_lc[i]]][[cnd]]$p[[i_nd[i]]], p_fa))
+			}
+		}
 	}
-	if (sminus)
-		return(call("-", expr))
-	else
-		return(expr)
+#browser()
+	# form final symbolic expression
+	# replace all i_lc by one product of fa_nd and lincomb of the reduced nds
+	fa_nd$num$b <- append(fa_nd$num$b, lc2expr(lc[i_lc]))
+	fa_nd$num$p <- append(fa_nd$num$p, 1)
+	lc <- c(list(fa_nd), lc[-i_lc])
+	return(lc2expr(lc))
 }
 
 `Simplify.-` <- function(expr)
@@ -209,6 +217,8 @@ Simplify_ <- function(expr)
 
 `Simplify.*` <- function(expr, div=FALSE)
 {
+#print(expr)
+#browser()
 	a <- expr[[2]]
 	b <- expr[[3]]
 	if (is.uminus(a)) {
@@ -254,31 +264,23 @@ Simplify_ <- function(expr)
 		}
 		# reduce numerics to only one factor
 		fa=list()
-		for (na in c("num", "den")) {
-			inu=if (length(nd[[na]]$b)) which(sapply(nd[[na]]$b, is.numeric) &
-				sapply(nd[[na]]$p, is.numeric)) else integer(0)
-			if (length(inu)) {
-				# the power of numerics are always 1 after Simplify_()
-				fa[[na]] <- prod(unlist(nd[[na]]$b[inu]))
-				# remove numerics
-				nd[[na]]$b <- nd[[na]]$b[-inu]
-				nd[[na]]$p <- nd[[na]]$p[-inu]
-			} else {
-				fa[[na]] <- 1
-			}
-			if (fa[[na]] < 0) {
-				sminus = !sminus
-				fa[[na]] <- -fa[[na]]
-			}
+		if (div) {
+			fa$num <- nd_a$fa$num*nd_b$fa$den
+			fa$den <- nd_a$fa$den*nd_b$fa$num
+		} else {
+			fa$num <- nd_a$fa$num*nd_b$fa$num
+			fa$den <- nd_a$fa$den*nd_b$fa$den
 		}
-		if (fa$num == 0) {
-			return(0)
-		}
-		if ((as.integer(fa$num) != fa$num ||
-				as.integer(fa$den) != fa$den) ||
-				fa$num == fa$den || fa$num == -fa$den) {
-			fa$num <- fa$num/fa$den
+		res <- fa$num/fa$den
+		if (as.integer(res) == res) {
+			fa$num <- res
 			fa$den <- 1
+		} else if (fa$den != 1) {
+			res <- fa$den/fa$num
+			if (as.integer(res) == res) {
+				fa$num <- 1
+				fa$den <- res
+			}
 		}
 		# group identical bases by adding their powers
 #browser()
@@ -316,8 +318,14 @@ Simplify_ <- function(expr)
 			if (length(iden)) {
 				# simplify power for this pair
 				ipair <- cbind(ipair, c(inum, iden))
-				nd$num$p[[inum]] <- Simplify_(call("-", nd$num$p[[inum]], nd$den$p[[iden]]))
-				nd$den$p[[iden]] <- 0
+				res <- Simplify_(call("-", nd$num$p[[inum]], nd$den$p[[iden]]))
+				if (res > 0) {
+					nd$num$p[[inum]] <- res
+					nd$den$p[[iden]] <- 0
+				} else {
+					nd$num$p[[inum]] <- 0
+					nd$den$p[[iden]] <- Simplify_(substitute(-res))
+				}
 			}
 		}
 		if (ncol(ipair) > 0) {
@@ -331,51 +339,18 @@ Simplify_ <- function(expr)
 			}
 		}
 #browser()
-		# form symbolic products
-		eprod <- list()
+		# remove power==0 terms
 		for (na in c("num", "den")) {
 			if (length(nd[[na]]$b) == 0)
 				next
-			# remove power==0 terms
 			ize=sapply(nd[[na]]$p, `==`, 0)
 			nd[[na]]$b <- nd[[na]]$b[!ize]
 			nd[[na]]$p <- nd[[na]]$p[!ize]
-			if (length(nd[[na]]$b) == 0)
-				next
-			for (i in seq_along(nd[[na]]$b)) {
-				p <- nd[[na]]$p[[i]]
-				term <- if (p == 1) nd[[na]]$b[[i]] else Simplify_(call("^", nd[[na]]$b[[i]], p))
-				if (is.null(eprod[[na]]))
-					eprod[[na]] <- term # start the sequence
-				else
-					eprod[[na]] <- call("*", eprod[[na]], term)
-			}
 		}
-		expr <- if (is.null(eprod$num)) 1 else eprod$num
-		if (!is.null(eprod$den)) {
-			expr <- call("/", expr, eprod$den)
-		}
-		# put numeric factor at first place
-		if (fa$num != 1 && fa$den != 1) {
-			# add to both num. and denom.
-			if (!is.null(eprod$den)) {
-				expr[[2]] <- call("*", fa$num, expr[[2]])
-				expr[[3]] <- call("*", fa$den, expr[[3]])
-			} else {
-				expr <- call("/", call("*", fa$num, expr), fa$den)
-			}
-		} else if (fa$num != 1) {
-			if (is.call(expr) && expr[[1]] == as.symbol("/") && expr[[2]] == 1)
-				expr[[2]] <- fa$num
-			else
-				expr <- call("*", fa$num, expr)
-		} else if (fa$den != 1) {
-			if (is.call(expr) && expr[[1]] == as.symbol("/"))
-				expr[[3]] <- call("*", fa$den, expr[[3]])
-			else
-				expr <- call("/", expr, fa$den)
-		}
-		return(if (sminus) substitute(-expr) else expr)
+		nd[["fa"]] <- fa
+		nd[["sminus"]] <- sminus
+		expr <- nd2expr(nd)
+		expr
 	}
 }
 `Simplify./` <- function(expr)
@@ -486,9 +461,39 @@ Simplify.abs <- function(expr) {
 	}
 	expr
 }
+Simplify.sign <- function(expr) {
+	if (is.uminus(expr[[2]])) {
+		expr[[2]] <- expr[[2]][[2]]
+		expr <- call("-", expr)
+	} else if (is.call(expr[[2]])) {
+		if (expr[[2]][[1]] == as.symbol("^")) {
+			p <- expr[[2]][[3]]
+			if (is.numeric(p) && p%%2 == 0)
+				expr <- 1
+		} else if (expr[[2]][[1]] == as.symbol("exp") || expr[[2]][[1]] == as.symbol("sqrt")) {
+			expr <- 1
+		}
+	}
+	expr
+}
+Simplify.if <- function(expr) {
+	cond <- expr[[2]]
+	if ((is.logical(cond) || is.numeric(cond)) && isTRUE(!!cond)) {
+		expr <- expr[[3]]
+	} else if (length(expr) == 4) {
+		if ((is.logical(cond) || is.numeric(cond)) && isTRUE(!cond)) {
+			expr <- expr[[4]]
+		} else if (format1(expr[[3]]) == format1(expr[[4]])) {
+			expr <- expr[[3]]
+		}
+	}
+	expr
+}
 
 Numden <- function(expr) {
 	# Return a list with "num" as numerator and "den" as denominator sublists.
+	# "fa" field is for numeric factors in "num" and "den" subfields.
+	# "sminus" is logical for applying or not "-" to the whole expression
 	# Each sublist regroups the language expressions which are not products neither
 	# divisions. The terms are decomposed in b^p sublists
 	if (is.uminus(expr)) {
@@ -498,13 +503,13 @@ Numden <- function(expr) {
 	} else if (is.uplus(expr)) {
 		Numden(expr[[2]])
 	} else if (is.symbol(expr)) {
-		if (is.numeric(expr)) {
-			sminus <- expr < 0
-			expr <- if (sminus) -expr else expr
-		} else {
-			sminus <- FALSE
-		}
-		list(num=list(b=expr, p=1), sminus=sminus)
+		list(num=list(b=list(expr), p=1),
+			sminus=FALSE,
+			fa=list(num=1, den=1))
+	} else if (is.numeric(expr)) {
+		sminus <- expr < 0
+		list(fa=list(num=if (sminus) -expr else expr, den=1),
+			sminus=sminus)
 	} else if (is.call(expr)) {
 		if (expr[[1]] == as.symbol("*")) {
 			# recursive call
@@ -512,26 +517,36 @@ Numden <- function(expr) {
 			b=Numden(expr[[3]])
 			list(num=list(b=c(a$num$b, b$num$b), p=c(a$num$p, b$num$p)),
 				den=list(b=c(a$den$b, b$den$b), p=c(a$den$p, b$den$p)),
-				sminus=xor(a$sminus, b$sminus))
+				sminus=xor(a$sminus, b$sminus),
+				fa=list(num=a$fa$num*b$fa$num, den=a$fa$den*b$fa$den))
 		} else if (expr[[1]] == as.symbol("/")) {
 			# recursive call
 			a=Numden(expr[[2]])
 			b=Numden(expr[[3]])
 			list(num=list(b=c(a$num$b, b$den$b), p=c(a$num$p, b$den$p)),
 				den=list(b=c(a$den$b, b$num$b), p=c(a$den$p, b$num$p)),
-				sminus=xor(a$sminus, b$sminus))
+				sminus=xor(a$sminus, b$sminus),
+				fa=list(num=a$fa$num*b$fa$den, den=a$fa$den*b$fa$num))
 		} else if (expr[[1]] == as.symbol("^")) {
 			if (expr[[3]] < 0) {
 				# make the power look positive
-				list(den=list(b=expr[[2]], p=if (is.numeric(expr[[3]])) -expr[[3]] else expr[[3]][[2]]), sminus=FALSE)
+				list(den=list(b=list(expr[[2]]), p=if (is.numeric(expr[[3]])) -expr[[3]] else expr[[3]][[2]]),
+					sminus=FALSE,
+					fa=list(num=1, den=1))
 			} else {
-				list(num=list(b=expr[[2]], p=expr[[3]]), sminus=FALSE)
+				list(num=list(b=list(expr[[2]]), p=expr[[3]]),
+					sminus=FALSE,
+					fa=list(num=1, den=1))
 			}
 		} else {
-			list(num=list(b=expr, p=1), sminus=FALSE)
+			list(num=list(b=list(expr), p=1),
+				sminus=FALSE,
+				fa=list(num=1, den=1))
 		}
 	} else {
-		list(num=list(b=expr, p=1), sminus=FALSE)
+		list(num=list(b=list(expr), p=1),
+			sminus=FALSE,
+			fa=list(num=1, den=1))
 	}
 }
 is.uminus <- function(e) {
@@ -547,47 +562,172 @@ is.unumeric <- function(e) {
 	return(is.numeric(e) || ((is.uminus(e) || is.uplus(e)) && is.unumeric(e[[2]])))
 }
 Lincomb <- function(expr) {
-	# decompose sum and diff in linear combibnation of num.coeff*item paires.
-	# For numerics, item is 1
-	# numerical factor is supposed to be at first place in products
-	if (is.uminus(expr)) {
-		a <- Lincomb(expr[[2]])
-		a$co=-a$co
-		a
-	} else if (is.uplus(expr)) {
-		Lincomb(expr[[2]])
-	} else if (is.symbol(expr)) {
-		list(it=list(expr), co=1)
-	} else if (is.numeric(expr)) {
-		list(co=expr, it=1)
-	} else if (is.call(expr)) {
+	# decompose expr in a list of product terms (cf Numden)
+	# the sign of each term is determined by the nd$sminus logical item.
+	if (is.call(expr) && length(expr) == 3) {
 		if (expr[[1]] == as.symbol("+")) {
 			# recursive call
-			a <- Lincomb(expr[[2]])
-			b <- Lincomb(expr[[3]])
-			list(co=c(a$co, b$co), it=c(a$it, b$it))
+			c(Lincomb(expr[[2]]), Lincomb(expr[[3]]))
 		} else if (expr[[1]] == as.symbol("-")) {
 			# recursive call
 			a <- Lincomb(expr[[2]])
 			b <- Lincomb(expr[[3]])
-			list(co=c(a$co, -b$co), it=c(a$it, b$it))
-		} else if (expr[[1]] == as.symbol("*") && is.numeric(expr[[2]])) {
-			list(co=expr[[2]], it=expr[[3]])
-		} else if (expr[[1]] == as.symbol("/")) {
-			if (is.numeric(expr[[2]]))
-				list(it=list(call("/", 1, expr[[3]])), co=expr[[2]])
-			else if (is.call(expr[[2]]) && expr[[2]][[1]] == as.symbol("*") && is.numeric(expr[[2]][[2]]))
-				list(it=list(call("/", expr[[2]][[3]]), expr[[3]]), co=expr[[2]][[2]])
-			else
-				list(co=1, it=list(expr))
+			# inverse the sign in b terms
+			b <- lapply(b, function(it) {it$sminus <- !it$sminus; it})
+			c(a, b)
 		} else {
-			list(co=1, it=list(expr))
+			list(Numden(expr))
 		}
 	} else {
-		list(co=1, it=list(expr))
+		list(Numden(expr))
 	}
 }
 
+# return an environement in wich stored subexpressions with
+# an index giving the position of each subexpression in the
+# whole statement st
+Leaves <- function(st, ind="1", res=new.env()) {
+	if (is.call(st)) {
+		res[[ind]] <- format1(st)
+		args <- as.list(st)[-1]
+		l <- lapply(seq_along(args), function(i) Leaves(args[[i]], paste(ind, i+1, sep="."), res))
+	}
+	return(res)
+}
+
+# replace repeated subexpressions by cached values
+Cache <- function(st, env=Leaves(st), prefix="") {
+	stch <- if (is.call(st)) as.character(st[[1]]) else ""
+	if (stch == "<-" || stch == "=") {
+		return(call("<-", st[[2]], Cache(st[[3]], prefix=paste(".", st[[1]], sep=""))))
+	} else if (stch == "{") {
+		return(as.call(c(list(st[[1]]), lapply(as.list(st)[-1], Cache))))
+	}
+	alva <- all.vars(st)
+	p <- grep(sprintf("^%s.e[0-9]+", prefix), alva, value=T)
+	if (length(p) > 0) {
+		prefix <- max(p)
+	}
+	ve <- unlist(as.list(env))
+	ta <- table(ve)
+	ta <- ta[ta > 1]
+	if (length(ta) == 0)
+		return(st)
+	e <- list() # will store the result code
+	alva <- list()
+	for (sub in names(sort(ta, decreasing=TRUE))) {
+		# get indexes for this subexpression
+		isubs <- names(which(ve == sub))
+		for (i in seq_along(isubs)) {
+			isub <- isubs[i]
+			subst <- parse(text=sprintf("st[[%s]]", gsub("\\.", "]][[", substring(isub, 3))))[[1]]
+			if (i == 1) {
+				esubst <- try(eval(subst), silent=TRUE)
+				if (inherits(esubst, "try-error"))
+					break # was already cached
+				# add subexpression to the final code
+				ie=length(e)+1
+				estr <- sprintf("%s.e%d", prefix, ie)
+				esub <- as.symbol(estr)
+				e[[ie]] <- call("<-", esub, esubst)
+				alva[[estr]] <- all.vars(esubst)
+			}
+			# replace subexpression in st by .eX
+			do.call(`<-`, list(subst, as.symbol("esub")))
+		}
+	}
+#browser()
+	alva[["end"]] <- all.vars(st)
+	# where .eX are used? If only once, develop, replace and remove it
+	wh <- lapply(seq_along(e), function(i) {
+		it=sprintf("%s.e%d", prefix, i)
+		which(sapply(alva, function(v) any(it == v)))
+	})
+	# the final touch
+	e[[ie+1]] <- st
+	dere <- sapply(wh, function(it) if (length(it) == 1 && names(it) != "end") it[[1]] else 0)
+	for (i in which(dere != 0)) {
+		idest <- dere[i]
+		li <- list()
+		li[[sprintf("%s.e%d", prefix, i)]] <- e[[i]][[3]]
+		e[[idest]][[3]] <- do.call("substitute", c(e[[idest]][[3]], list(li)))
+	}
+	e <- c(list(as.symbol("{")), e[which(!dere)], e[length(e)])
+	return(as.call(e))
+}
+nd2expr <- function(nd, sminus=NULL) {
+	# form symbolic products
+	# if sminus is not null, use it instead of the nd's one
+	if (length(nd) == 0)
+		return(0)
+	eprod <- list()
+	for (na in c("num", "den")) {
+		if (length(nd[[na]]$b) == 0)
+			next
+		# alphabetic order for bases, symbols first, then calls
+		for (i in order(sapply(nd[[na]]$b, is.call), sapply(nd[[na]]$b, format1))) {
+			p <- nd[[na]]$p[[i]]
+			if (p == 0)
+				next
+			term <- if (p == 1) nd[[na]]$b[[i]] else Simplify_(call("^", nd[[na]]$b[[i]], p))
+			if (is.null(eprod[[na]]))
+				eprod[[na]] <- term # start the sequence
+			else
+				eprod[[na]] <- call("*", eprod[[na]], term)
+		}
+	}
+	expr <- if (is.null(eprod$num)) 1 else eprod$num
+	if (!is.null(eprod$den)) {
+		expr <- call("/", expr, eprod$den)
+	}
+	# put numeric factor at first place
+	fa=nd$fa
+	if (fa$num != 1 && fa$den != 1) {
+		# add to both num. and denom.
+		if (!is.null(eprod$den)) {
+			expr[[2]] <- call("*", fa$num, expr[[2]])
+			expr[[3]] <- call("*", fa$den, expr[[3]])
+		} else {
+			expr <- call("/", call("*", fa$num, expr), fa$den)
+		}
+	} else if (fa$num != 1) {
+		if (is.call(expr) && expr[[1]] == as.symbol("/") && expr[[2]] == 1)
+			expr[[2]] <- fa$num
+		else
+			expr <- call("*", fa$num, expr)
+	} else if (fa$den != 1) {
+		if (is.call(expr) && expr[[1]] == as.symbol("/"))
+			expr[[3]] <- call("*", fa$den, expr[[3]])
+		else
+			expr <- call("/", expr, fa$den)
+	}
+	expr <- if ((!is.null(sminus) && sminus) || (is.null(sminus) && nd$sminus)) substitute(-expr) else expr
+#print(sprintf("nd->%s", format1(expr)))
+	return(expr)
+}
+lc2expr <- function(lc) {
+	# form symbolic sum and diff form a list of nds
+	# separate in positive and negative
+	smin <- sapply(lc, "[[", "sminus")
+	epos <- lapply(lc[which(!smin)], nd2expr)
+	eneg <- lapply(lc[which(smin)], nd2expr, sminus=FALSE)
+	if (length(epos) == 0)
+		return(if (length(eneg) == 0) 0 else substitute(-li2sum(eneg)))
+	else
+		return(if (length(eneg) == 0) li2sum(epos) else call("-", li2sum(epos), li2sum(eneg)))
+}
+li2sum <- function(li) {
+	# form a long sum of expressions from the list li
+	if (length(li) == 0)
+		0
+	else if (length(li) == 1)
+		li[[1]]
+	else if (length(li) == 2)
+		call("+", li[[1]], li[[2]])
+	else
+		call("+", li[[1]], li2sum(li[-1]))
+}
+
 simplifications <- new.env()
 
 assign("+", `Simplify.+`, envir=simplifications)
@@ -600,3 +740,5 @@ assign("log", `Simplify.log`, envir=simplifications)
 assign("logb", `Simplify.log`, envir=simplifications)
 assign("sqrt", `Simplify.sqrt`, envir=simplifications)
 assign("abs", `Simplify.abs`, envir=simplifications)
+assign("sign", `Simplify.sign`, envir=simplifications)
+assign("if", `Simplify.if`, envir=simplifications)
